(defpackage :fits
  (:use :common-lisp)
  (:export :leer-fits :data :cabecera))

(in-package :fits)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;
;; Code for reading FITS images

(defclass imagen-fits ()
  ((cabecera :initarg :cabecera
	     :accessor cabecera)
   (data     :initarg :data
	     :accessor data)))

(declaim (inline trim-spaces-subseq))
(defun trim-spaces-subseq (cadena inicio fin)
  (string-trim " " (subseq cadena inicio fin))) 

(defun leer-cabecera (archivo &optional (imprimir))
  "Reads the header of the FITS file.
   TODO: -check if EXTEND or END keywords are present. If not -> bad, bad file...
         -If EXTEND -> read another 36 lines."
  (let ((linea-vector (make-array 80 :element-type '(unsigned-byte 8)))
	(cabecera     (make-hash-table :test 'equal))
	(fin          nil)
	(extend       nil))
    (do ()
	(fin)
      (dotimes (i 36)
	(read-sequence linea-vector archivo)
	(let* ((linea (map 'string (function code-char) linea-vector))
	       (llave (trim-spaces-subseq linea 0 8))
	       (valor (trim-spaces-subseq linea 9 30)))
	  (cond
	    ((zerop (length llave))) ; Nothing in the line...
	    ((string= llave "END")     (setf fin t))           ; This is good!
	    ((string= llave "EXTEND")  (setf extend t))        ; Then I should read 36 more lines...
	    (t
	     (setf (gethash llave cabecera) (if (find #\' valor)
						valor
						(read-from-string valor)))))
	  (when imprimir (format t "~a ~a~%" llave valor)))))
    cabecera))

(defun crear-imagen-fits (cabecera)
  (let (ejes)
    (dotimes (naxis (gethash "NAXIS" cabecera))
      (push (gethash (format nil "~a~a" "NAXIS" (+ naxis 1)) cabecera) ejes))
    (make-instance 'imagen-fits
		   :cabecera cabecera
		   :data   (make-array (nreverse ejes)))))

(defmethod leer-data ((imagen imagen-fits) archivo)
  "Reads the pixels data.
   TODO: be able to read n-dimensional data."
  (let* ((data   (data imagen))
	 (ancho  (array-dimension data 0))
	 (alto   (array-dimension data 1))
	 (bscale (or (gethash "BSCALE" (cabecera imagen)) 1))
	 (bzero  (or (gethash "BZERO"  (cabecera imagen)) 0))
	 (bitpix (gethash "BITPIX" (cabecera imagen))))
    (dotimes (j alto)
      (dotimes (i ancho)
	(let ((aux 0))
	  (loop for k downfrom (- bitpix 8) to 0 by 8 do
	       (setf (ldb (byte 8 k) aux) (read-byte archivo)))
	  (setf (aref data i j) (+ bzero (* aux bscale))))))))

(defun leer-fits (nombre &optional imprimir)
  (with-open-file (archivo nombre :element-type '(unsigned-byte 8))
    (let ((imagen-fits (crear-imagen-fits (leer-cabecera archivo imprimir))))
      (leer-data imagen-fits archivo)
      imagen-fits)))


